// Copyright (C) 2006-2013 Xilinx, Inc.  All rights reserved
// $Id: GridMap.h,v 1.16.206.2 2013/05/10 00:04:05 jcorbett Exp $
//
// Map locations on a 2D grid for adjacency testing.
//
// This template must be parameterized with a type T that has a method
// with signature
//    void T::Visit(T* neighbor)
// that will be called for every non-null unordered pair of adjacent
// entries in the grid map.
//
// The typical Visit method would compare this with neighbor and
// record constraint violations if necessary.
//
// Each (x, y) location may be assigned a non-null value at most
// once. In this application we never need to assign a grid cell more
// than once and having the check caught a bug.
//
// Note: It is the caller's responsibility to insure that pointers
// passed to MarkPoint are valid when GridMap::SweepAllAdjacent() or
// GridMap::SweepVerticallyAdjacent() are invoked. The GridMap class
// does not deallocate pointers passed to it.
//
// R#check_package_pins, R#check_interconnect_routing,
// R#check_direct_routing, R#check_logic_tiles

#ifndef Iv_GridMap_h
#define Iv_GridMap_h

#include <vector>
#include "XOS/Portability/export/Port_Type.h"

#include "main.h"

template <typename GridCell>
class GridMap {
public:
   // Construct a grid map of given dimensions and given grid
   // storage. The vector reference is passed in so the caller may
   // conveniently handle deallocation.
   GridMap(unsigned int x, unsigned int y, std::vector<GridCell*>& grid)
      : xSize(x), ySize(y), grid(grid)
      {
         ASSERT(x != 0 && y != 0);
         grid.resize(x * y);
      }
   virtual ~GridMap() {}

   // Mark a point in the two dimensional grid with a non-null grid
   // cell pointer.
   void MarkPoint(unsigned int x, unsigned int y, GridCell* gridCell) {
      ASSERT(x < xSize && y < ySize); // (x, y) in bounds.
      ASSERT(!grid[Index(x, y)]);     // Assign at most once.
      ASSERT(gridCell);               // Can't store null.
      // Recall grid.at(i) will throw if i is out of bounds, but grid[i]
      // won't.
      grid.at(Index(x, y)) = gridCell;
   }
   // Return the pointer in cell (x, y).
   GridCell* At(Port::UINT32 x, Port::UINT32 y) {
      if (InGrid(x, y)) {
         return grid.at(Index(x, y));
      } else {
         return 0;
      }
   }

   // Check grid coordinates for validity.
   bool InGrid(Port::UINT32 x, Port::UINT32 y) {
      return x < ColumnCount() && y < RowCount();
   }

   Port::UINT32 RowCount() {
      return ySize;
   }

   Port::UINT32 ColumnCount() {
      return xSize;
   }
   // Invoke the Visit method on every unordered pair of grid cells
   // that are both occupied.
   //
   // Since adjacency is a symmetric relation, it suffices to check half
   // of the eight neighbors.
   //
   //   +---------+---------+---------+
   //   |         |         |         |
   //   |         |         |         |
   //   +---------+---------+---------+
   //   |         |  here   | right   |
   //   |         |  (x,y)  | (x+1,y) |
   //   +---------+---------+---------+
   //   | below   |  below  | below   |
   //   | left    |         | right   |
   //   |(x-1,y+1)| (x,y+1) |(x+1,y+1)|
   //   +---------+---------+---------+
   //
   void SweepAllAdjacent() {
      for (unsigned int x(0); x < xSize; ++x) {
         for (unsigned int y(0); y < ySize; ++y) {
            GridCell* here(grid[Index(x, y)]);
            if (here) {
               if (x + 1 < xSize) {
                  GridCell* right(grid[Index(x + 1, y)]);
                  if (right) {
                     here->Visit(right);
                  }
                  if (y + 1 < ySize) {
                     GridCell* belowRight(grid[Index(x + 1, y + 1)]);
                     if (belowRight) {
                        here->Visit(belowRight);
                     }
                  }
               }
               if (y + 1 < ySize) {
                  GridCell* below(grid[Index(x, y + 1)]);
                  if (below) {
                     here->Visit(below);
                  }
               }
               if (0 < x && y + 1 < ySize) {
                  GridCell* belowLeft(grid[Index(x - 1, y + 1)]);
                  if (belowLeft) {
                     here->Visit(belowLeft);
                  }
               }
            }
         }
      }
   }
   // Invoke Visit on every vertically adjacent unordered pair.  Note:
   // we are concerned with adjacency of int_interconnect
   // tiles. Even/odd pairs of vertically adjacent tiles share an
   // interconnect tile, so in IOB coordinates, 'adjacent' means Y
   // values separated by 1, 2, if the lesser coordinate is odd;
   // and 1, 2, or 3 otherwise.
   //
   //   IOB_X0Y5 \____ red INT_INTERCONNECT
   //   IOB_X0Y4 /
   //
   //   IOB_X0Y3 \____ unused INT_INTERCONNECT
   //   IOB_X0Y2 /
   //
   //   IOB_X0Y1 \____ black INT_INTERCONNECT
   //   IOB_X0Y0 /
   //   ...
   //
   void SweepVerticallyAdjacent() {
      for (unsigned int y(0); y < ySize - 1; ++y) {
         unsigned int gapMax(y % 2 ? 2 : 3);
         for (unsigned int gap(1); gap <= gapMax; ++gap) {
            unsigned int yAbove(y + gap);
            if (yAbove < ySize) {
               for (unsigned int x(0); x < xSize; ++x) {
                  GridCell* here(grid[Index(x, y)]);
                  if (here) {
                     GridCell* above(grid[Index(x, yAbove)]);
                     if (above) {
                        here->Visit(above);
                     }
                  }
               }
            }
         }
      }
   }
   unsigned int RowCount() const    { return ySize; }
   unsigned int ColumnCount() const { return xSize; }

   // Do the address arithmetic that lets us treat a one dimensional
   // vector as a two dimensional matrix.
   virtual unsigned int Index(unsigned int x, unsigned int y) {
      ASSERT(x < xSize && y < ySize);
      return xSize * y + x;
   }
protected:
   const unsigned int xSize;
   const unsigned int ySize;
   std::vector<GridCell*>& grid;
};

#endif
